# nmake macros for building Windows 32-Bit apps
!INCLUDE <win32.mak>

DIST_DIR=..\..\..\bin\$(TYPE)

# If the rpc library is not included in the standard lib path
# you have to give the path to it here.
RPCLIB=..\..\librpc\obj\$(TYPE)

# If the rpc include directory is not included in the standard path
# you have to give the path to it here.
RPCINC=..\..\..\include

# If you want to use the new multithreaded server feature uncomment this line
# We do not fully support this option in this Makefile.
# MT = 1

# If rpcgen is not in your path, you should set the directory where it lives
RPCBIN=..\..\rpcgen\obj\$(TYPE)

RPCGEN=rpcgen

!UNDEF TYPE

!IFDEF MT
MT_cvars=-DMULTITHREAD $(cvarsmt)
MT_libs=$(conlibsmt)
!ELSE
MT_cvars=$(cvars)
MT_libs=$(conlibs)
!ENDIF

!IFDEF NODEBUG
!MESSAGE Doing release targets.
TYPE=Release
CFLAGS=$(cflags) $(MT_cvars) $(cdebug) -DNDEBUG -D_NDEBUG -I"$(RPCINC)"
LFLAGS=$(lflags) $(ldebug) -libpath:"$(RPCLIB)"
!ELSE
!MESSAGE Doing debug targets.
TYPE=Debug
CFLAGS=$(cflags) $(MT_cvars) $(cdebug) -D_DEBUG -DDEBUG -I"$(RPCINC)"
LFLAGS=$(lflags) $(ldebug) -libpath:"$(RPCLIB)"
!ENDIF

!IFDEF RPCBIN
PATH=$(RPCBIN);$(PATH)
!ENDIF

OUTDIR=obj\$(TYPE)

.c{$(OUTDIR)}.obj:
	$(cc) $(CFLAGS) /Fo$@ /c $<

.SUFFIXES: .x

.x.h:
	$(RPCGEN) $<

.x.c:
	$(RPCGEN) $<

RPC_PROC_C = $(DEMO)_proc.c
RPC_CLNT_C = r$(DEMO).c

RPC_PROC_OBJ = $(OUTDIR)\$(RPC_PROC_C:.c=.obj)
RPC_CLNT_OBJ = $(OUTDIR)\$(RPC_CLNT_C:.c=.obj)

RPC_X = $(DEMO).x
RPC_X_H = $(RPC_X:.x=.h)
RPC_X_XDR = $(RPC_X:.x=_xdr.c)
RPC_X_SVC = $(RPC_X:.x=_svc.c)
RPC_X_CLNT = $(RPC_X:.x=_clnt.c)
RPC_X_GEN = $(RPC_X_H) $(RPC_X_XDR) $(RPC_X_SVC) $(RPC_X_CLNT)

RPC_X_XDR_OBJ = $(OUTDIR)\$(RPC_X_XDR:.c=.obj)
RPC_X_SVC_OBJ = $(OUTDIR)\$(RPC_X_SVC:.c=.obj)
RPC_X_CLNT_OBJ = $(OUTDIR)\$(RPC_X_CLNT:.c=.obj)
RPC_X_OBJS = $(RPC_X_XDR_OBJ) $(RPC_X_SVC_OBJ) $(RPC_X_CLNT_OBJ)

RPC_CLNT_EXE = $(RPC_CLNT_OBJ:.obj=.exe)
RPC_SVC_EXE = $(RPC_X_SVC_OBJ:.obj=.exe)


CLNT_OBJS = $(RPC_CLNT_OBJ) $(RPC_X_XDR_OBJ) $(RPC_X_CLNT_OBJ)
SVC_OBJS = $(RPC_PROC_OBJ) $(RPC_X_XDR_OBJ) $(RPC_X_SVC_OBJ)

TARGETS= $(RPC_CLNT_EXE) $(RPC_SVC_EXE)

all: $(OUTDIR) $(TARGETS)

$(DIST_DIR):
	mkdir $(DIST_DIR)

dist: all $(DIST_DIR)
	@for %%a in ( $(TARGETS) ) do @copy %%a $(DIST_DIR)\.

clean:
	-@for %%a in ( $(RPC_X_GEN) $(RPC_CLNT_OBJ) $(RPC_PROC_OBJ) $(RPC_X_OBJS) $(TARGETS) ) do @if exist %%a erase %%a
	-@for %%a in ( $(OUTDIR) obj ) do @if exist %%a rmdir %%a
	@echo Cleaned $(OUTDIR)

help:
	@echo Define NODEBUG to get release targets
	@echo Otherwise, you will get debug targets
	@echo nmake all - builds binaries (default)
	@echo nmake dist - copies binaries
	@echo nmake clean - removes binaries
	@echo nmake help - prints this message

$(OUTDIR):
	@if not exist $(OUTDIR) mkdir $(OUTDIR)

$(RPC_CLNT_EXE): $(CLNT_OBJS)
	$(link) $(LFLAGS) -out:$@ $** $(MT_libs) oncrpc.lib

$(RPC_SVC_EXE): $(SVC_OBJS)
	$(link) $(LFLAGS) -out:$@ $** $(MT_libs) oncrpc.lib

$(RPC_CLNT_OBJ): $(RPC_CLNT_C) $(RPC_X_H)
$(RPC_PROC_OBJ): $(RPC_PROC_C) $(RPC_X_H)
$(RPC_X_XDR_OBJ): $(RPC_X_XDR)
$(RPC_X_SVC_OBJ): $(RPC_X_SVC)
$(RPC_X_CLNT_OBJ): $(RPC_X_CLNT)
$(RPC_X_GEN): $(RPC_X)
