/* @(#)dir_proc.c	2.1 88/08/02 4.0 RPCSRC */
/*
 * dir_proc.c: remote readdir implementation
 */
#include <rpc/rpc.h>
#include "dir.h"

#ifndef WIN32

#include <sys/dir.h>

extern int errno;
extern char *malloc();
extern char *strcpy();

#else /* WIN32 */

#include <windows.h>

struct direct {
    char d_name[MAX_PATH];
};

static WIN32_FIND_DATA wfd;

typedef struct {
    HANDLE h;
    WIN32_FIND_DATA data;
} DIR;

DIR*
opendir(
    char* dirname
    )
{
    DIR *d = (DIR*)malloc(sizeof(DIR));
    d->h = FindFirstFile(dirname, &d->data);
    if (d->h == INVALID_HANDLE_VALUE) {
	errno = (int)GetLastError();
	free(d);
	d = 0;
    }
    return d;
}

void
closedir(
    DIR* d
    )
{
    FindClose(d->h);
    free(d);
}

struct direct*
readdir(
    DIR* d
    )
{
    static struct direct de;
    if (d->h != INVALID_HANDLE_VALUE) {
	strcpy(de.d_name, d->data.cFileName);
	if (!FindNextFile(d->h, &d->data)) {
	    FindClose(d->h);
	    d->h = INVALID_HANDLE_VALUE;
	}
	return &de;
    } else {
	return 0;
    }
}

#endif /* WIN32 */

readdir_res *
readdir_1(dirname)
	nametype *dirname;
{
	DIR *dirp;
	struct direct *d;
	namelist nl;
	namelist *nlp;
	static readdir_res res; /* must be static! */
	
	/*
	 * Open directory
	 */
	dirp = opendir(*dirname);
	if (dirp == NULL) {
		res.errno = errno;
		return (&res);
	}

	/*
	 * Free previous result
	 */
	xdr_free(xdr_readdir_res, &res);

	/*
	 * Collect directory entries
	 */
	nlp = &res.readdir_res_u.list;
	while (d = readdir(dirp)) {
		nl = *nlp = (namenode *) malloc(sizeof(namenode));
		nl->name = malloc(strlen(d->d_name)+1);
		strcpy(nl->name, d->d_name);
		nlp = &nl->next;
	}
	*nlp = NULL;

	/*
	 * Return the result
	 */
	res.errno = 0;
	closedir(dirp);
	return (&res);
}
