/*********************************************************************
 * RPC for the Windows NT Operating System
 * 1993 by Martin F. Gergeleit
 * Users may use, copy or modify Sun RPC for the Windows NT Operating 
 * System according to the Sun copyright below.
 *
 * RPC for the Windows NT Operating System COMES WITH ABSOLUTELY NO 
 * WARRANTY, NOR WILL I BE LIABLE FOR ANY DAMAGES INCURRED FROM THE 
 * USE OF. USE ENTIRELY AT YOUR OWN RISK!!!
 *********************************************************************/

/*
 * Copyright (c) 1984 by Sun Microsystems, Inc.
 */

/*
 * portmap.c, Implements the program,version to port number mapping for
 * rpc.
 */

/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */

#ifndef WIN32
#pragma error /* will hopefully cause us to die */
#endif

#include <rpc/rpc.h>
#include <rpc/pmap_prot.h>
#include <stdio.h>

#include "portmap.h"
#include "service.h"

void get_myaddress(struct sockaddr_in *);

static void reg_service(struct svc_req *, SVCXPRT *);
static void callit(struct svc_req *, SVCXPRT *);

struct pmaplist *pmaplist;
static SVCXPRT *ludpxprt, *ltcpxprt;
static int removesvc = 0;
static int installsvc = 0;
static int service = 0;
static int debugging = 0;
static int verbose = 0;

char optstr[128];
char* progname = 0;

int getopt(int, char **, char *);

void
usage()
{
	fprintf(stderr, 
		"usage: %s [-i] [-r] [-s] [-d] [-v]\n"
		"\t-i\tinstall service\n"
		"\t-r\tremove service\n"
		"\t-s\trun as service\n"
		"\t-d\tdebugging output\n"
		"\t-v\tverbose output\n", progname);
	exit(1);
}

void
main(
    int argc,
    char **argv
    )
{
	progname = argv[0];
	if (!options(argc, argv)) {
		usage();
		exit(1);
	}
	if (removesvc && installsvc) {
		usage();
		exit(1);
	} else if (removesvc) {
		RemoveService();
	} else if (installsvc) {
		InstallService();
	} else if (service) {
		run_service();
	} else {
		if (rpc_nt_init() || !init()) {
			report_error("could not init");
			exit(1);
		}
		report_info("portmap running");
		run(); /* blocks forever, actually... */
		rpc_nt_exit();
		report_info("portmap exited");
	}
}

bool_t
options(argc, argv)
	int argc;
	char **argv;
{
	int c;
	char* p = optstr;

	while ((c = getopt(argc, argv, "risdv")) != -1) {
		switch (c) {

		case 'r':
			removesvc = 1;
			break;
		case 'i':
			if (!service && 
			    ((p-optstr) < sizeof(optstr))) *(p++) = 's';
			installsvc = service = 1;
			break;
		case 's':
			if (!service && 
			    ((p-optstr) < sizeof(optstr))) *(p++) = c;
			service = 1;
			break;
		case 'd':
			if (!debugging && 
			    ((p-optstr) < sizeof(optstr))) *(p++) = c;
			debugging = 1;
			break;
		case 'v':
			if (!verbose && 
			    ((p-optstr) < sizeof(optstr))) *(p++) = c;
			verbose = 1;
			break;
		default:
			*(p++) = 0;
			return FALSE;
		}
	}

	*(p++) = 0;
	return TRUE;
}

bool_t
init()
{
	SVCXPRT *xprt;
	int sock, lsock;
	struct sockaddr_in addr, laddr;
	int on = 1;
	int len = sizeof(struct sockaddr_in);
	register struct pmaplist *pml;

	/* WIN32 - XXX - Add options code here... */


	memset(&addr, 0, sizeof(addr));
	addr.sin_addr.s_addr = 0;
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	addr.sin_port = htons(PMAPPORT);

	memset(&laddr, 0, sizeof(laddr));
	laddr.sin_addr.s_addr = 0;
	laddr.sin_family = AF_INET;
	laddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
	laddr.sin_port = htons(PMAPPORT);

	if ((sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == INVALID_SOCKET) {
		report_error("cannot create udp socket");
		return FALSE;
	}
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));
	if (bind(sock, (struct sockaddr *)&addr, len) != 0) {
		report_error("cannot bind udp");
		return FALSE;
	}

	if ((xprt = svcudp_create(sock)) == (SVCXPRT *)NULL) {
		report_error("couldn't do udp_create");
		return FALSE;
	}

	if ((lsock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP)) == INVALID_SOCKET) {
		report_error("cannot create udp socket");
		return FALSE;
	}
	setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));
	if (bind(lsock, (struct sockaddr *)&laddr, len) != 0) {
		report_error("cannot bind udp");
		return FALSE;
	}

	if ((ludpxprt = svcudp_create(lsock)) == (SVCXPRT *)NULL) {
		report_error("couldn't do udp_create");
		return FALSE;
	}

	/* make an entry for ourself */
	pml = (struct pmaplist *)malloc((u_int)sizeof(struct pmaplist));
	pml->pml_next = 0;
	pml->pml_map.pm_prog = PMAPPROG;
	pml->pml_map.pm_vers = PMAPVERS;
	pml->pml_map.pm_prot = IPPROTO_UDP;
	pml->pml_map.pm_port = PMAPPORT;
	pmaplist = pml;

	if ((sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == INVALID_SOCKET) {
		report_error("cannot create tcp socket");
		return FALSE;
	}
	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));
	if (bind(sock, (struct sockaddr *)&addr, len) != 0) {
		report_error("cannot bind tcp");
		return FALSE;
	}
	if ((xprt = svctcp_create(sock, RPCSMALLMSGSIZE, RPCSMALLMSGSIZE))
	    == (SVCXPRT *)NULL) {
		report_error("couldn't do tcp_create");
		return FALSE;
	}

	if ((lsock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) == INVALID_SOCKET) {
		report_error("cannot create tcp socket");
		return FALSE;
	}
	setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, (void *)&on, sizeof(on));
	if (bind(lsock, (struct sockaddr *)&laddr, len) != 0) {
		report_error("cannot bind tcp");
		return FALSE;
	}
	if ((ltcpxprt = svctcp_create(lsock, RPCSMALLMSGSIZE,
	    RPCSMALLMSGSIZE)) == (SVCXPRT *)NULL) {
		report_error("couldn't do tcp_create");
		return FALSE;
	}

	/* make an entry for ourself */
	pml = (struct pmaplist *)malloc((u_int)sizeof(struct pmaplist));
	pml->pml_map.pm_prog = PMAPPROG;
	pml->pml_map.pm_vers = PMAPVERS;
	pml->pml_map.pm_prot = IPPROTO_TCP;
	pml->pml_map.pm_port = PMAPPORT;
	pml->pml_next = pmaplist;
	pmaplist = pml;

	(void)svc_register(xprt, PMAPPROG, PMAPVERS, reg_service, FALSE);
	return TRUE;
}

void
run(void)
{
	svc_run();
}

static struct pmaplist *
find_service(prog, vers, prot)
	u_long prog;
	u_long vers;
	u_long prot;
{
	register struct pmaplist *hit = NULL;
	register struct pmaplist *pml;
	char str[256];

	for (pml = pmaplist; pml != NULL; pml = pml->pml_next) {
		if ((pml->pml_map.pm_prog != prog) ||
			(pml->pml_map.pm_prot != prot))
			continue;
		hit = pml;
		if (pml->pml_map.pm_vers == vers)
		    break;
	}

        if (verbose) {  /* MJJ: 14-11-96 added this condition */
		if (hit != NULL)
			sprintf(str, "found program %d, version %d with protocol %d, port %d",
				prog, vers, hit->pml_map.pm_prot, hit->pml_map.pm_port);
		else
			sprintf(str, "cannot find program %d, version %d", prog, vers);
		report_info(str);
        }
        
	return (hit);
}

/* 
 * 1 OK, 0 not
 */
void
reg_service(rqstp, xprt)
	struct svc_req *rqstp;
	SVCXPRT *xprt;
{
	struct pmap reg;
	struct pmaplist *pml, *prevpml, *fnd;
	struct sockaddr_in *fromsin;
	int ans, port;
	caddr_t t;
	char str[1024];
	
	fromsin = svc_getcaller(xprt);

	switch (rqstp->rq_proc) {

	case PMAPPROC_NULL:
		/*
		 * Null proc call
		 */
		if (!svc_sendreply(xprt, (xdrproc_t)xdr_void, NULL) &&
		    debugging)
			report_info("svc_sendreply");
		break;

	case PMAPPROC_SET:
		/*
		 * Set a program,version to port mapping
		 */
		if (xprt != ltcpxprt && xprt != ludpxprt) {
			sprintf(str,
				"non-local set attempt (might be from %s)",
				inet_ntoa(fromsin->sin_addr));
			report_info(str);
			svcerr_noproc(xprt);
			return;
		}
		if (!svc_getargs(xprt, xdr_pmap, (caddr_t)&reg)) {
			svcerr_decode(xprt);
			return;
		}
		/*
		 * check to see if already used
		 * find_service returns a hit even if
		 * the versions don't match, so check for it
		 */
		ans = 0;
		fnd = find_service(reg.pm_prog, reg.pm_vers, reg.pm_prot);
		if (fnd && fnd->pml_map.pm_vers == reg.pm_vers) {
			if (fnd->pml_map.pm_port == reg.pm_port)
				ans = 1;
		} else {
			/* 
			 * add to list
			 */
			pml = (struct pmaplist *)
				    malloc((u_int)sizeof(struct pmaplist));
			pml->pml_map = reg;
			pml->pml_next = pmaplist;
			pmaplist = pml;
			ans = 1;
		}
		if (!svc_sendreply(xprt, xdr_long, (caddr_t)&ans) && debugging)
			report_info("svc_sendreply");
		break;

	case PMAPPROC_UNSET:
		/*
		 * Remove a program,version to port mapping.
		 */
		if (xprt != ltcpxprt && xprt != ludpxprt) {
			sprintf(str,
				"non-local set attempt (might be from %s)",
				inet_ntoa(fromsin->sin_addr));
			report_info(str);
			svcerr_noproc(xprt);
			return;
		}
		if (!svc_getargs(xprt, xdr_pmap, (caddr_t)&reg)) {
			svcerr_decode(xprt);
			return;
		}
		ans = 0;
		for (prevpml = NULL, pml = pmaplist; pml != NULL; ) {
			if ((pml->pml_map.pm_prog != reg.pm_prog) ||
				(pml->pml_map.pm_vers != reg.pm_vers)) {
				/* both pml & prevpml move forwards */
				prevpml = pml;
				pml = pml->pml_next;
				continue;
			}
			/* found it; pml moves forward, prevpml stays */
			ans = 1;
			t = (caddr_t)pml;
			pml = pml->pml_next;
			if (prevpml == NULL)
				pmaplist = pml;
			else
				prevpml->pml_next = pml;
			free(t);
		}
		if (!svc_sendreply(xprt, xdr_long, (caddr_t)&ans) && debugging)
			report_info("svc_sendreply");
		break;

	case PMAPPROC_GETPORT:
		/*
		 * Lookup the mapping for a program,version and return its port
		 */
		if (!svc_getargs(xprt, xdr_pmap, (caddr_t)&reg)) {
			svcerr_decode(xprt);
			return;
		}
		fnd = find_service(reg.pm_prog, reg.pm_vers, reg.pm_prot);
		if (fnd)
			port = fnd->pml_map.pm_port;
		else
			port = 0;
		if (!svc_sendreply(xprt, xdr_long, (caddr_t)&port) &&
		    debugging)
			report_info("svc_sendreply");
		break;

	case PMAPPROC_DUMP:
		/*
		 * Return the current set of mapped program,version
		 */
		if (!svc_getargs(xprt, (xdrproc_t)xdr_void, NULL)) {
			svcerr_decode(xprt);
			return;
		}
		if (!svc_sendreply(xprt, xdr_pmaplist, (caddr_t)&pmaplist) && 
		    debugging)
			report_info("svc_sendreply");
		break;

	case PMAPPROC_CALLIT:
		/*
		 * Calls a procedure on the local machine.  If the requested
		 * procedure is not registered this procedure does not return
		 * error information!!
		 * This procedure is only supported on rpc/udp and calls via 
		 * rpc/udp.  It passes null authentication parameters.
		 */
		callit(rqstp, xprt);
		/* WIN32 - XXX - make it non-blocking */
		break;

	default:
		svcerr_noproc(xprt);
		break;
	}
}


/*
 * Stuff for the rmtcall service
 */
#define ARGSIZE 9000

typedef struct encap_parms {
	u_long arglen;
	char *args;
} encap_parms_t;

static bool_t
xdr_encap_parms(xdrs, epp)
	XDR *xdrs;
	struct encap_parms *epp;
{

	return (xdr_bytes(xdrs, &(epp->args), &(epp->arglen), ARGSIZE));
}

typedef struct rmtcallargs {
	u_long	rmt_prog;
	u_long	rmt_vers;
	u_long	rmt_port;
	u_long	rmt_proc;
	struct encap_parms rmt_args;
} rmtcallargs_t;

static bool_t
xdr_rmtcall_args(xdrs, cap)
	register XDR *xdrs;
	register struct rmtcallargs *cap;
{

	/* does not get a port number */
	if (xdr_u_long(xdrs, &(cap->rmt_prog)) &&
	    xdr_u_long(xdrs, &(cap->rmt_vers)) &&
	    xdr_u_long(xdrs, &(cap->rmt_proc))) {
		return (xdr_encap_parms(xdrs, &(cap->rmt_args)));
	}
	return (FALSE);
}

static bool_t
xdr_rmtcall_result(xdrs, cap)
	register XDR *xdrs;
	register struct rmtcallargs *cap;
{
	if (xdr_u_long(xdrs, &(cap->rmt_port)))
		return (xdr_encap_parms(xdrs, &(cap->rmt_args)));
	return (FALSE);
}

/*
 * only worries about the struct encap_parms part of struct rmtcallargs.
 * The arglen must already be set!!
 */
static bool_t
xdr_opaque_parms(xdrs, cap)
	XDR *xdrs;
	struct rmtcallargs *cap;
{

	return (xdr_opaque(xdrs, cap->rmt_args.args, cap->rmt_args.arglen));
}

/*
 * This routine finds and sets the length of incoming opaque paraters
 * and then calls xdr_opaque_parms.
 */
static bool_t
xdr_len_opaque_parms(xdrs, cap)
	register XDR *xdrs;
	struct rmtcallargs *cap;
{
	register u_int beginpos, lowpos, highpos, currpos, pos;

	beginpos = lowpos = pos = xdr_getpos(xdrs);
	highpos = lowpos + ARGSIZE;
	while ((int)(highpos - lowpos) >= 0) {
		currpos = (lowpos + highpos) / 2;
		if (xdr_setpos(xdrs, currpos)) {
			pos = currpos;
			lowpos = currpos + 1;
		} else {
			highpos = currpos - 1;
		}
	}
	xdr_setpos(xdrs, beginpos);
	cap->rmt_args.arglen = pos - beginpos;
	return (xdr_opaque_parms(xdrs, cap));
}

/*
 * Call a remote procedure service
 * This procedure is very quiet when things go wrong.
 * The proc is written to support broadcast rpc.  In the broadcast case,
 * a machine should shut-up instead of complain, less the requestor be
 * overrun with complaints at the expense of not hearing a valid reply ...
 */
static void
callit(rqstp, xprt)
	struct svc_req *rqstp;
	SVCXPRT *xprt;
{
	struct rmtcallargs a;
	struct pmaplist *pml;
	u_short port;
	struct sockaddr_in me;
	int so = -1, dontblock = 1;
	CLIENT *client;
	struct authunix_parms *au = (struct authunix_parms *)rqstp->rq_clntcred;
	struct timeval timeout;
	char buf[ARGSIZE];

	timeout.tv_sec = 5;
	timeout.tv_usec = 0;
	a.rmt_args.args = buf;
	if (!svc_getargs(xprt, xdr_rmtcall_args, (caddr_t)&a))
	    return;
	if ((pml = find_service(a.rmt_prog, a.rmt_vers, IPPROTO_UDP)) == NULL)
	    return;

	port = (u_short) pml->pml_map.pm_port;
	get_myaddress(&me);
	me.sin_port = htons(port);

	/* Avoid implicit binding to reserved port by clntudp_create() */
	so = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (so < 0) {
		svcerr_systemerr(xprt);
		return;
	}
	if (ioctlsocket(so, FIONBIO, &dontblock) < 0) {
		closesocket(so);
		svcerr_systemerr(xprt);
		return;
	}
	client = clntudp_create(&me, a.rmt_prog, a.rmt_vers, timeout, &so);
	if (client != (CLIENT *)NULL) {
		if (rqstp->rq_cred.oa_flavor == AUTH_UNIX) {
			client->cl_auth = authunix_create(au->aup_machname,
			   au->aup_uid, au->aup_gid, au->aup_len, au->aup_gids);
		}
		a.rmt_port = (u_long)port;
		if (clnt_call(client, a.rmt_proc, xdr_opaque_parms, 
			      (caddr_t)&a, xdr_len_opaque_parms, (caddr_t)&a, 
			      timeout) == RPC_SUCCESS) {
			svc_sendreply(xprt, xdr_rmtcall_result, (caddr_t)&a);
		}
		AUTH_DESTROY(client->cl_auth);
		clnt_destroy(client);
	}
	(void)closesocket(so);
}

void
report_error(
    const char* msg
    )
{
	if (service) {
		service_report(msg, EVENTLOG_ERROR_TYPE);
	} else {
		fprintf(stderr, "%s: %s (%d)\n", 
			progname, msg, GetLastError());
	}
}

void
report_info(
    const char* msg
    )
{
	if (service) {
		service_report(msg, EVENTLOG_INFORMATION_TYPE);
	} else {
		printf("%s: %s\n", progname, msg);
	}
}
