/*********************************************************************
 * RPC for the Windows NT Operating System
 * 1993 by Martin F. Gergeleit
 *
 * RPC for the Windows NT Operating System COMES WITH ABSOLUTELY NO
 * WARRANTY, NOR WILL I BE LIABLE FOR ANY DAMAGES INCURRED FROM THE
 * USE OF. USE ENTIRELY AT YOUR OWN RISK!!!
 *********************************************************************/
/*
*  Added on suggestion from Martin Johnson (MJJ)
*
*  The -v flag means VERBOSE, i.e. generate NT Application
*  Event Log messages in Portmap.c/find_service() when
*  clients try to connect.
*
*  To enable VERBOSE mode the service has to be started manually from 
*  Control Panel/Services, with something in the "Startup Parameters"
*  field.  (Note: "NET START PORTMAP -V" won't work as the -V gets
*  thrown away.)
*/

#include <rpc/rpc.h>
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "portmap.h"
#include "service.h"

void
service_main(
    DWORD dwArgc,
    char **argv
    );

LPCSTR PORTMAP_SVC_NAME = "Portmap";

VOID
WINAPI
service_ctrl(
    DWORD dwCtrlCode
    );

BOOL
ReportStatusToSCMgr(
    DWORD dwCurrentState,
    DWORD dwWin32ExitCode,
    DWORD dwCheckPoint,
    DWORD dwWaitHint
    );

void
StopPortmapService(
    LPCSTR lpszMsg
    );

HANDLE                  hServDoneEvent = NULL;
SERVICE_STATUS          ssStatus;
SERVICE_STATUS_HANDLE   sshStatusHandle;
DWORD                   dwGlobalErr;
DWORD                   TID = 0;
HANDLE                  threadHandle = NULL;
HANDLE                  pipeHandle;


void
run_service(
    )
{
    SERVICE_TABLE_ENTRY dispatchTable[] = {
        { PORTMAP_SVC_NAME, (LPSERVICE_MAIN_FUNCTION)service_main },
        { NULL, NULL }
    };
    if (!StartServiceCtrlDispatcher(dispatchTable)) {
        StopPortmapService("cannot start service control dispatcher");
    }
}


void
service_main(
    DWORD dwArgc,
    char **argv
    )
{
    DWORD dwWait;

    sshStatusHandle = RegisterServiceCtrlHandler(PORTMAP_SVC_NAME,
                                                 (LPHANDLER_FUNCTION)
                                                 service_ctrl);

    if (!sshStatusHandle)
        goto exit_portmap;

    ssStatus.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
    ssStatus.dwServiceSpecificExitCode = 0;

    if (!ReportStatusToSCMgr(SERVICE_START_PENDING, NO_ERROR, 1, 3000))
        goto exit_portmap;

    hServDoneEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

    if (hServDoneEvent == (HANDLE)NULL)
        goto exit_portmap;


    if (!ReportStatusToSCMgr(SERVICE_START_PENDING, NO_ERROR, 2, 3000))
        goto exit_portmap;

    if (rpc_nt_init() != 0) {
	goto exit_portmap;
    }

    if (!init()) {
        StopPortmapService("cannot init");
	goto exit_portmap;
    }

    threadHandle = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)run, 
                                NULL, 0, &TID);

    if (!threadHandle)
        goto exit_portmap;

    if (!ReportStatusToSCMgr(SERVICE_RUNNING, NO_ERROR, 0, 0))
        goto exit_portmap;

    report_info("service running");
    dwWait = WaitForSingleObject(hServDoneEvent, INFINITE);
    TerminateThread(threadHandle, 0);

exit_portmap:

    report_info("service terminated");

    rpc_nt_exit();

    if (hServDoneEvent != NULL)
        CloseHandle(hServDoneEvent);

    if (sshStatusHandle != 0)
        (VOID)ReportStatusToSCMgr(SERVICE_STOPPED, dwGlobalErr, 0, 0);

    return;
}


VOID
WINAPI
service_ctrl(
    DWORD dwCtrlCode
    )
{
    DWORD  dwState = SERVICE_RUNNING;

    switch(dwCtrlCode) {

        case SERVICE_CONTROL_PAUSE:

            if (ssStatus.dwCurrentState == SERVICE_RUNNING) {
                SuspendThread(threadHandle);
                dwState = SERVICE_PAUSED;
            }
            break;

        case SERVICE_CONTROL_CONTINUE:

            if (ssStatus.dwCurrentState == SERVICE_PAUSED) {
                ResumeThread(threadHandle);
                dwState = SERVICE_RUNNING;
            }
            break;

        case SERVICE_CONTROL_STOP:

            dwState = SERVICE_STOP_PENDING;

            ReportStatusToSCMgr(SERVICE_STOP_PENDING, NO_ERROR, 1, 3000);

            SetEvent(hServDoneEvent);
            return;

        case SERVICE_CONTROL_INTERROGATE:
            break;

        default:
            break;

    }

    ReportStatusToSCMgr(dwState, NO_ERROR, 0, 0);
}


BOOL
ReportStatusToSCMgr(
    DWORD dwCurrentState,
    DWORD dwWin32ExitCode,
    DWORD dwCheckPoint,
    DWORD dwWaitHint
    )
{
    BOOL fResult;

    if (dwCurrentState == SERVICE_START_PENDING)
        ssStatus.dwControlsAccepted = 0;
    else
        ssStatus.dwControlsAccepted = SERVICE_ACCEPT_STOP |
            SERVICE_ACCEPT_PAUSE_CONTINUE;

    ssStatus.dwCurrentState = dwCurrentState;
    ssStatus.dwWin32ExitCode = dwWin32ExitCode;
    ssStatus.dwCheckPoint = dwCheckPoint;

    ssStatus.dwWaitHint = dwWaitHint;

    if (!(fResult = SetServiceStatus(sshStatusHandle, &ssStatus)))
        StopPortmapService("cannot set service status");

    return fResult;
}


void
service_report(
    LPCSTR lpszMsg,
    WORD wEventLogInfoType
    )
{
    CHAR chMsg[1024];
    LPSTR  lpszStrings[1];

    DWORD dwGlobalErr = GetLastError();
    HANDLE hEventSource = RegisterEventSource(NULL,
                                              PORTMAP_SVC_NAME);

    if (!hEventSource)
        return;

    switch(wEventLogInfoType) { 
    case EVENTLOG_ERROR_TYPE:
        sprintf(chMsg, "error: %s (%d)", lpszMsg, dwGlobalErr);
        break;
    default:
        sprintf(chMsg, "info: %s", lpszMsg);
    }

    lpszStrings[0] = chMsg;

    ReportEvent(hEventSource,
                wEventLogInfoType,
                0,
                0,
                NULL,
                1,
                0,
                lpszStrings,
                NULL);

    DeregisterEventSource(hEventSource);
}

void
StopPortmapService(
    LPCSTR lpszMsg
    )
{
    report_error(lpszMsg);
    SetEvent(hServDoneEvent);
}

#define MAX_CMD 1024

void
InstallService()
{
    SC_HANDLE manager;
    SC_HANDLE service;

    char cmd[MAX_CMD];
    char* cmdp = cmd;
    int optlen = strlen(optstr); /* has to contain at least service option */
    /* XXX - we really should assert optlen */
    optlen += 3; /* space, -, NULL */

    *(cmdp++) = '"'; /* bquote */

    if (!GetModuleFileName(NULL, cmdp, MAX_CMD - optlen - 1 /* bquote */)) {
	printf("error: cannot get name of port mapper service binary "
               "(0x%08X)\n",
	       GetLastError());
	exit(1);
    }

    if (optlen) {
	strcat(cmd, "\" -"); /* equote, space, - */
	strcat(cmd, optstr); /* now the options */
    }

    manager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!manager) {
	printf("error: cannot open service control manager (0x%08X)\n", 
	       GetLastError());
	exit(1);
    }

    service = CreateService(manager,
			    PORTMAP_SVC_NAME,
			    PORTMAP_SVC_NAME,
			    SERVICE_ALL_ACCESS,
			    SERVICE_WIN32_OWN_PROCESS,
			    SERVICE_DEMAND_START,
			    SERVICE_ERROR_NORMAL,
			    cmd,
			    NULL,
			    NULL,
			    NULL,
			    NULL,
			    NULL);

    if (!service) {
        printf("error: cannot create port mapper service (0x%08X)\n",
	       GetLastError());
        exit(1);
    }

    printf("port mapper service successfully installed\n");
    CloseServiceHandle(service);
    CloseServiceHandle(manager);
}

void
RemoveService()
{
    SC_HANDLE manager;
    SC_HANDLE service;

    manager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!manager) {
	printf("error: cannot open service control manager (0x%08X)\n", 
	       GetLastError());
	exit(1);
    }

    service = OpenService(manager, PORTMAP_SVC_NAME, SERVICE_ALL_ACCESS);

    if (!service) {
        printf("error: cannot open port mapper service (0x%08X)\n", 
	       GetLastError());
	exit(1);
    }

    if (!DeleteService(service)) {
        printf("error: cannot delete port mapper service (0x%08X)\n", 
               GetLastError());
	exit(1);
    }

    CloseServiceHandle(service);
    CloseServiceHandle(manager);
    printf("port mapper service successfully deleted\n");
}
